@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -33,15 +33,22 @@
 	  [ source_to_html/3		% +Source, +OutStream, +Options
 	  ]).
 :- use_module(library(option)).
+:- use_module(library(debug)).
 :- use_module(library(lists)).
+:- use_module(library(prolog_colour)).
 :- use_module(doc_colour).
 :- use_module(doc_html).
 :- use_module(doc_wiki).
 :- use_module(doc_modes).
 :- use_module(doc_process).
-:- use_module(library('http/html_write')).
+:- use_module(library(http/html_write)).
+:- use_module(library(http/http_path)).
 :- use_module(library(prolog_xref)).
 
+:- meta_predicate
+	source_to_html(+, +, :).
+
+
 /** <module> HTML source pretty-printer
 
 This module colourises Prolog  source  using   HTML+CSS  using  the same
@@ -51,37 +58,74 @@ cross-reference based technology as used by PceEmacs.
 @author Jan Wielemaker
 */
 
+:- predicate_options(source_to_html/3, 3,
+		     [ format_comments(boolean),
+		       header(boolean),
+		       skin(callable),
+		       stylesheets(list),
+		       title(atom)
+		     ]).
+
+
 :- thread_local
 	lineno/0,			% print line-no on next output
 	nonl/0.				% previous tag implies nl (block level)
 
-%%	source_to_html(+In:filename, +Out, +Options) is det.
+%%	source_to_html(+In:filename, +Out, :Options) is det.
 %
-%	Colourise Prolog source as HTML. The idea is to first create a
-%	sequence of fragments and then to apply these to the code.
+%	Colourise Prolog source as HTML. The idea   is to first create a
+%	sequence of fragments and  then  to   apply  these  to the code.
+%	Options are passed to the following predicates:
 %
-%	@param In	A filename
+%	  * print_html_head/2
+%	  * print_html_footer/2.
+%	  * html_fragments/6
+%
+%	@param In	A filename.  Can also be an abstract name,
+%			which is subject to library(prolog_source)
+%			abstract file handling. See
+%			prolog_open_source/2.  Note that this cannot
+%			be a stream as we need to read the file three
+%			times: (1) xref, (2) assign colours and (3)
+%			generate HTML.
 %	@param Out	Term stream(Stream) or file-name specification
 
-source_to_html(Src, stream(Out), Options) :- !,
+source_to_html(Src, stream(Out), MOptions) :- !,
+	meta_options(is_meta, MOptions, Options),
+	(   option(title(_), Options)
+	->  HeadOptions = Options
+	;   file_base_name(Src, Title),
+	    HeadOptions = [title(Title)|Options]
+	),
 	retractall(lineno),		% play safe
 	retractall(nonl),		% play safe
 	colour_fragments(Src, Fragments),
-	open(Src, read, In),
-	asserta(user:message_hook(_,_,_), Ref),
-	call_cleanup((file_base_name(Src, Base),
-		      print_html_head(Out, [title(Base), Options]),
-		      html_fragments(Fragments, In, Out, [], State, Options),
-		      copy_rest(In, Out, State, State1),
-		      pop_state(State1, Out, In)),
-		     (erase(Ref),
-		      close(In))),
+	setup_call_cleanup(
+	    ( open_source(Src, In),
+	      asserta(user:thread_message_hook(_,_,_), Ref)
+	    ),
+	    ( print_html_head(Out, HeadOptions),
+	      html_fragments(Fragments, In, Out, [], State, Options),
+	      copy_rest(In, Out, State, State1),
+	      pop_state(State1, Out, In)
+	    ),
+	    ( erase(Ref),
+	      close(In)
+	    )),
 	print_html_footer(Out, Options).
 source_to_html(Src, FileSpec, Options) :-
 	absolute_file_name(FileSpec, OutFile, [access(write)]),
-	open(OutFile, write, Out, [encoding(utf8)]),
-	call_cleanup(source_to_html(Src, stream(Out), Options),
-		     close(Out)).
+	setup_call_cleanup(
+	    open(OutFile, write, Out, [encoding(utf8)]),
+	    source_to_html(Src, stream(Out), Options),
+	    close(Out)).
+
+open_source(Id, Stream) :-
+	prolog:xref_open_source(Id, Stream), !.
+open_source(File, Stream) :-
+	open(File, read, Stream).
+
+is_meta(skin).
 
 %%	print_html_head(+Out:stream, +Options) is det.
 %
@@ -98,13 +142,20 @@ source_to_html(Src, FileSpec, Options) :-
 %
 %		* format_comments(Bool)
 %		If =true= (default), format structured comments.
+%
+%		* skin(Closure)
+%		Called using call(Closure, Where, Out), where Where
+%		is one of =header= or =footer=.  These calls are made
+%		just after opening =body= and before closing =body=.
 
 print_html_head(Out, Options) :-
 	option(header(true), Options, true), !,
 	option(title(Title), Options, 'Prolog source'),
-	option(stylesheets(Sheets), Options, ['pllisting.css', 'pldoc.css']),
+	http_absolute_location(pldoc_resource('pldoc.css'), PlDocCSS, []),
+	http_absolute_location(pldoc_resource('pllisting.css'), PlListingCSS, []),
+	option(stylesheets(Sheets), Options, [PlListingCSS, PlDocCSS]),
 	format(Out,
-	       '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" \
+	       '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" \c
 	       "http://www.w3.org/TR/html4/strict.dtd">~n~n', []),
 	format(Out, '<html>~n', []),
 	format(Out, '  <head>~n', []),
@@ -112,14 +163,23 @@ print_html_head(Out, Options) :-
 	forall(member(Sheet, Sheets),
 	       format(Out, '    <link rel="stylesheet" type="text/css" href="~w">~n', [Sheet])),
 	format(Out, '  </head>~n', []),
-	format(Out, '<body>~n', []).
-print_html_head(_, _).
+	format(Out, '<body>~n', []),
+	skin_hook(Out, header, Options).
+print_html_head(Out, Options) :-
+	skin_hook(Out, header, Options).
 
 print_html_footer(Out, Options) :-
 	option(header(true), Options, true), !,
+	skin_hook(Out, footer, Options),
 	format(Out, '~N</body>~n', []),
 	format(Out, '</html>', []).
-print_html_footer(_, _).
+print_html_footer(Out, Options) :-
+	skin_hook(Out, footer, Options).
+
+skin_hook(Out, Where, Options) :-
+	option(skin(Skin), Options),
+	call(Skin, Where, Out), !.
+skin_hook(_, _, _).
 
 
 %%	html_fragments(+Fragments, +In, +Out, +State, +Options) is det.
@@ -190,6 +250,7 @@ start_fragment(Class, _, Out, State, [span(class(SpanClass))|State]) :-
 
 end_fragment(_, _, [nop|State], State) :- !.
 end_fragment(Out, In, [span(class(directive))|State], State) :- !,
+	copy_full_stop(In, Out),
 	format(Out, '</span>', []),
 	(   peek_code(In, 10),
 	    \+ nonl
@@ -225,6 +286,22 @@ copy_to(In, End, Out, State, [pre(class(listing))|State]) :-
 	assert(lineno),
 	write_codes(Codes, Line, Out).
 
+%%	copy_full_stop(+In, +Out) is det.
+%
+%	Copy upto and including the .
+
+copy_full_stop(In, Out) :-
+	get_code(In, C0),
+	copy_full_stop(C0, In, Out).
+
+copy_full_stop(0'., _, Out) :- !,
+	put_code(Out, 0'.).
+copy_full_stop(C, In, Out) :-
+	put_code(Out, C),
+	get_code(In, C2),
+	copy_full_stop(C2, In, Out).
+
+
 %%	delete_leading_white_lines(+CodesIn, -CodesOut, +LineIn, -Line) is det.
 %
 %	Delete leading white lines. Used  after structured comments. The
@@ -275,6 +352,11 @@ all_spaces([H|T]) :-
 copy_to(In, End, Out) :-
 	line_count(In, Line),
 	read_to(In, End, Codes),
+	(   debugging(htmlsrc)
+	->  length(Codes, Count),
+	    debug(htmlsrc, 'Copy ~D chars: ~s', [Count, Codes])
+	;   true
+	),
 	write_codes(Codes, Line, Out).
 
 read_to(In, End, Codes) :-
@@ -336,13 +418,15 @@ copy_rest(In, Out, State0, State) :-
 %	Read the next N codes from In as a list of codes. If N < 0, read
 %	upto the end of stream In.
 
-read_n_codes(_, 0, []) :- !.
+read_n_codes(_, N, Codes) :-
+	N =< 0, !,
+	Codes = [].
 read_n_codes(In, N, Codes) :-
 	get_code(In, C0),
 	read_n_codes(N, C0, In, Codes).
 
-read_n_codes(1, C, _, [C]) :- !.
 read_n_codes(_, -1, _, []) :- !.
+read_n_codes(1, C, _, [C]) :- !.
 read_n_codes(N, C, In, [C|T]) :-
 	get_code(In, C2),
 	N2 is N - 1,
@@ -366,7 +450,7 @@ element_clause(element(Term, Tag, CSS)) :-
 	element_tag(Term, Tag).
 
 span_term(Classification, Class) :-
-	prolog_src_style(Classification, _Style),
+	syntax_colour(Classification, _Attributes),
 	css_class(Classification, Class).
 
 css_class(Class, Class) :-
