@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2010, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
 			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
@@ -20,7 +18,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -32,11 +30,12 @@
 
 :- module(pldoc_index,
 	  [ doc_for_dir/2,		% +Dir, +Options
-	    dir_index/4,		% +Dir, +Options, //
-	    object_summaries/5,		% +Objs, +Section, +Options, //
-	    file_index_header/4,	% +File, +Options, //
-	    doc_links/4,		% +Directory, +Options, //
-	    doc_file_href/2,		% +File, -/doc/...
+	    dir_index//2,		% +Dir, +Options, //
+	    object_summaries//3,	% +Objs, +Section, +Options, //
+	    file_index_header//2,	% +File, +Options, //
+	    doc_links//2,		% +Directory, +Options, //
+	    doc_file_href/2,		% +File, -HREF
+	    places_menu//1,		% +Dir, //
 	    source_directory/1		% ?Directory
 	  ]).
 :- use_module(doc_process).
@@ -50,12 +49,47 @@
 :- use_module(library(readutil)).
 :- use_module(library(url)).
 :- use_module(library(option)).
+:- use_module(library(lists)).
 :- use_module(library(doc_http)).
 :- include(hooks).
 
 /** <module> Create indexes
 */
 
+:- predicate_options(dir_index//2, 2,
+		     [ directory(atom),
+		       edit(boolean),
+		       files(list),
+		       members(list),
+		       qualify(boolean),
+		       title(atom),
+		       if(oneof([true,loaded])),
+		       recursive(boolean),
+		       secref_style(oneof([number, title, number_title])),
+		       pass_to(doc_links/4, 2)
+		     ]).
+:- predicate_options(doc_links//2, 2,
+		     [ files(list),
+		       pass_to(pldoc_search:search_form/3, 1)
+		     ]).
+:- predicate_options(file_index_header//2, 2,
+		     [ directory(any),
+		       files(list),
+		       qualify(boolean),
+		       secref_style(oneof([number, title, number_title])),
+		       pass_to(pldoc_html:edit_button/4, 2),
+		       pass_to(pldoc_html:source_button/4, 2)
+		     ]).
+:- predicate_options(object_summaries//3, 3,
+		     [ edit(boolean),
+		       files(list),
+		       module(atom),
+		       public(list),
+		       qualify(boolean),
+		       secref_style(oneof([number, title, number_title]))
+		     ]).
+:- predicate_options(doc_for_dir/2, 2, [pass_to(dir_index/4, 2)]).
+
 %%	doc_for_dir(+Dir, +Options) is det.
 %
 %	Write summary index for all files  in   Dir  to  Out. The result
@@ -69,25 +103,48 @@ doc_for_dir(DirSpec, Options) :-
 			     access(read)
 			   ],
 			   Dir),
-	file_base_name(Dir, Base),
-	reply_html_page(pldoc(dir_index),
-			title(Base),
-			\dir_index(Dir, Options)).
+	(   option(title(Title), Options)
+	->  true
+	;   file_base_name(Dir, Title)
+	),
+	doc_write_page(
+	    pldoc(dir_index),
+	    title(Title),
+	    \dir_index(Dir, Options),
+	    Options).
+
+:- html_meta doc_write_page(+, html, html, +).
+
+doc_write_page(Style, Head, Body, Options) :-
+	option(files(_), Options), !,
+	phrase(page(Style, Head, Body), HTML),
+	print_html(HTML).
+doc_write_page(Style, Head, Body, _) :-
+	reply_html_page(Style, Head, Body).
 
 
 %%	dir_index(+Dir, +Options)//
 %
 %	Create an index for all Prolog files appearing in Dir or in
-%	any directory contained in Dir.
+%	any directory contained in Dir.  Options:
+%
+%	  * members(+Members)
+%	  Documented members.  See doc_files.pl
+%	  * title(+Title)
+%	  Title to use for the index page
 
 dir_index(Dir, Options) -->
 	{ dir_source_files(Dir, Files0, Options),
 	  sort(Files0, Files),
-	  atom_concat(Dir, '/index.html', File),
+	  maplist(ensure_doc_objects, Files),
+	  directory_file_path(Dir, 'index.html', File),
 	  b_setval(pldoc_file, File)	% for predref
 	},
-	html([ \doc_links(Dir, Options),
+	html([ \doc_resources(Options),
+	       \doc_links(Dir, Options),
 	       \dir_header(Dir, Options),
+	       \subdir_links(Dir, Options),
+	       h2(class([wiki,plfiles]), 'Prolog files'),
 	       table(class(summary),
 		     \file_indices(Files, [directory(Dir)|Options])),
 	       \dir_footer(Dir, Options)
@@ -97,59 +154,100 @@ dir_index(Dir, Options) -->
 %
 %	Create a list of source-files to be documented as part of Dir.
 
-dir_source_files(DirSpec, Files, _Options) :-
-	absolute_file_name(DirSpec, Dir,
-			   [ file_type(directory),
-			     access(read)
-			   ]),
-	findall(F, source_file_in_dir(Dir, F), Files).
+dir_source_files(_, Files, Options) :-
+	option(members(Members), Options), !,
+	findall(F, member(file(F,_Doc), Members), Files).
+dir_source_files(Dir, Files, Options) :-
+	directory_source_files(Dir, Files, Options).
 
-source_file_in_dir(Dir, File) :-
-	source_file(File),
-	file_directory_name(File, Dir).
+%%	subdir_links(+Dir, +Options)// is det.
+%
+%	Create links to subdirectories
+
+subdir_links(Dir, Options) -->
+	{ option(members(Members), Options),
+	  findall(SubDir, member(directory(SubDir, _, _, _), Members), SubDirs),
+	  SubDirs \== []
+	},
+	html([ h2(class([wiki,subdirs]), 'Sub directories'),
+	       table(class(subdirs),
+		     \subdir_link_rows(SubDirs, Dir))
+	     ]).
+subdir_links(_, _) --> [].
+
+subdir_link_rows([], _) --> [].
+subdir_link_rows([H|T], Dir) -->
+	subdir_link_row(H, Dir),
+	subdir_link_rows(T, Dir).
+
+subdir_link_row(Dir, From) -->
+	{ directory_file_path(Dir, 'index.html', Index),
+	  relative_file_name(Index, From, Link),
+	  file_base_name(Dir, Base)
+	},
+	html(tr(td(a([class(subdir), href(Link)], ['[dir] ', Base])))).
 
 %%	dir_header(+Dir, +Options)// is det.
 %
-%	Create header for directory
+%	Create header for directory.  Options:
+%
+%	  * readme(File)
+%	  Include File as introduction to the directory header.
 
-dir_header(Dir, _Options) -->
-	wiki_file(Dir, readme), !.
-dir_header(Dir, _Options) -->
-	{ file_base_name(Dir, Base)
+dir_header(Dir, Options) -->
+	wiki_file(Dir, readme, Options), !.
+dir_header(Dir, Options) -->
+	{ (   option(title(Title), Options)
+	  ->  true
+	  ;   file_base_name(Dir, Title)
+	  )
 	},
-	html(h1(class=dir, Base)).
+	html(h1(class=dir, Title)).
 
 %%	dir_footer(+Dir, +Options)// is det.
 %
 %	Create footer for directory. The footer contains the =TODO= file
-%	if provided.
+%	if provided.  Options:
+%
+%	  * todo(File)
+%	  Include File as TODO file in the footer.
 
-dir_footer(Dir, _Options) -->
-	wiki_file(Dir, todo), !.
+dir_footer(Dir, Options) -->
+	wiki_file(Dir, todo, Options), !.
 dir_footer(_, _) -->
 	[].
 
-%%	wiki_file(+Dir, +Type)// is semidet.
+%%	wiki_file(+Dir, +Type, +Options)// is semidet.
 %
 %	Include text from a Wiki text-file.
 
-wiki_file(Dir, Type) -->
-	{ wiki_file_type(Type, Base),
-	  atomic_list_concat([Dir, /, Base], File),
-	  access_file(File, read), !,
-	  read_file_to_codes(File, String, []),
+wiki_file(Dir, Type, Options) -->
+	{ (   Opt =.. [Type,WikiFile],
+	      option(Opt, Options)
+	  ->  true
+	  ;   directory_files(Dir, Files),
+	      member(File, Files),
+	      wiki_file_type(Type, Pattern),
+	      downcase_atom(File, Pattern),
+	      directory_file_path(Dir, File, WikiFile)
+	  ),
+	  access_file(WikiFile, read), !,
+	  read_file_to_codes(WikiFile, String, []),
 	  wiki_codes_to_dom(String, [], DOM)
 	},
 	pldoc_html:html(DOM).
 
 %%	wiki_file_type(+Category, -File) is nondet.
+%
+%	Declare file pattern names that are included for README and TODO
+%	for a directory. Files are matched case-insensitively.
 
-wiki_file_type(readme, 'README').
-wiki_file_type(readme, 'README.TXT').
-wiki_file_type(readme, 'README.txt').
-wiki_file_type(todo,   'TODO').
-wiki_file_type(todo,   'TODO.TXT').
-wiki_file_type(todo,   'TODO.txt').
+wiki_file_type(readme, 'readme').
+wiki_file_type(readme, 'readme.md').
+wiki_file_type(readme, 'readme.txt').
+wiki_file_type(todo,   'todo').
+wiki_file_type(todo,   'todo.md').
+wiki_file_type(todo,   'todo.txt').
 
 %%	file_indices(+Files, +Options)// is det.
 %
@@ -167,9 +265,7 @@ file_indices([H|T], Options) -->
 %	Create an index for File.
 
 file_index(File, Options) -->
-	{ Pos = File:_Line,
-	  findall(doc(Obj,Pos,Summary),
-		  doc_comment(Obj, Pos, Summary, _), Objs0),
+	{ doc_summaries(File, Objs0),
 	  module_info(File, ModuleOptions, Options),
 	  doc_hide_private(Objs0, Objs1, ModuleOptions),
 	  sort(Objs1, Objs)
@@ -178,6 +274,23 @@ file_index(File, Options) -->
 	     | \object_summaries(Objs, File, ModuleOptions)
 	     ]).
 
+doc_summaries(File, Objects) :-
+	xref_current_source(FileSpec),
+	xref_option(FileSpec, comments(collect)), !,
+	Pos = File:0,
+	findall(doc(Obj,Pos,Summary),
+		xref_doc_summary(Obj, Pos, Summary), Objects).
+doc_summaries(File, Objects) :-
+	Pos = File:_Line,
+	findall(doc(Obj,Pos,Summary),
+		doc_comment(Obj, Pos, Summary, _), Objects).
+
+xref_doc_summary(M:Name/Arity, File:_, Summary) :-
+	xref_comment(File, Head, Summary, _Comment),
+	xref_module(File, Module),
+	strip_module(Module:Head, M, Plain),
+	functor(Plain, Name, Arity).
+
 %%	file_index_header(+File, +Options)// is det.
 %
 %	Create an entry in a summary-table for File.
@@ -185,21 +298,26 @@ file_index(File, Options) -->
 file_index_header(File, Options) -->
 	prolog:doc_file_index_header(File, Options), !.
 file_index_header(File, Options) -->
-	{ file_base_name(File, Label),
-	  doc_file_href(File, HREF, Options),
-	  ButtonOptions = [button_height(16)|Options]
+	{ (   option(directory(Dir), Options),
+	      directory_file_path(Dir, Label, File)
+	  ->  true
+	  ;   file_base_name(File, Label)
+	  ),
+	  doc_file_href(File, HREF, Options)
 	},
 	html(tr(th([colspan(3), class(file)],
 		   [ span(style('float:left'), a(href(HREF), Label)),
 		     \file_module_title(File),
 		     span(style('float:right'),
-			  [ \source_button(File, ButtonOptions),
-			    \edit_button(File, ButtonOptions)
+			  [ \source_button(File, Options),
+			    \edit_button(File, Options)
 			  ])
 		   ]))).
 
 file_module_title(File) -->
-	{ module_property(M, file(File)),
+	{ (   module_property(M, file(File))
+	  ;   xref_module(File, M)
+	  ),
 	  doc_comment(M:module(Title), _, _, _)
 	}, !,
 	html([&(nbsp), ' -- ', Title]).
@@ -284,6 +402,10 @@ tag_pub_priv([H|T0], [Tag-H|T], Options) :-
 %
 %	@tbd	Hacky interface.  Do we demand Summary to be in Wiki?
 
+object_summary(doc(Obj, _Pos, _Summary), wiki, Options) --> !,
+	html(tr(class(wiki),
+		[ td(colspan(3), \object_ref(Obj, Options))
+		])).
 object_summary(doc(Obj, _Pos, Summary), _Section, Options) --> !,
 	(   { string_to_list(Summary, Codes),
 	      wiki_codes_to_dom(Codes, [], DOM0),
@@ -324,15 +446,17 @@ doc_links(_Directory, Options) -->
 	{ option(files(_), Options), !
 	}.
 doc_links(Directory, Options) -->
-	{   Directory == ''
-	->  working_directory(Dir, Dir)
-	;   Dir = Directory
+	{   (   Directory == ''
+	    ->  working_directory(Dir, Dir)
+	    ;   Dir = Directory
+	    ),
+	    option(html_resources(Resoures), Options, pldoc)
 	},
-	html([ \html_requires(pldoc),
+	html([ \html_requires(Resoures),
 	       div(class(navhdr),
 		   [ div(class(jump),
 			  div([ \places_menu(Dir),
-				\version
+				\plversion
 			      ])),
 		     div(class(search), \search_form(Options)),
 		     br(clear(right))
@@ -344,14 +468,22 @@ doc_links(Directory, Options) -->
 %
 %	Prolog version
 
-version -->
+plversion -->
 	{ current_prolog_flag(version_data, swi(Major, Minor, Patch, _))
-	},
+	}, !,
 	html(a([ class(prolog_version),
 		 href('http://www.swi-prolog.org')
 	       ],
 	       [' SWI-Prolog ', Major, '.', Minor, '.', Patch])).
 
+plversion -->
+	{ current_prolog_flag(version_data, yap(Major, Minor, Patch, _))
+	},
+	html(a([ class(prolog_version),
+		 href('http://www.dcc.fc.up.pt/~vsc')
+	       ],
+	       [' YAP ', Major, '.', Minor, '.', Patch])).
+
 
 %%	places_menu(Current)// is det
 %
@@ -363,13 +495,17 @@ places_menu(Dir) -->
 	{ findall(D, source_directory(D), List),
 	  sort(List, Dirs)
 	},
-	html(form([ action(location_by_id(pldoc_dir))
+	html(form([ action(location_by_id(go_place))
 		  ],
 		  [ input([type(submit), value('Go')]),
-		    select(name(dir),
-			   \source_dirs(Dirs, Dir))
+		    select(name(place),
+			   \packs_source_dirs(Dirs, Dir))
 		  ])).
 
+packs_source_dirs(Dirs, Dir) -->
+	packs_link,
+	source_dirs(Dirs, Dir).
+
 source_dirs([], _) -->
 	[].
 source_dirs([H|T], WD) -->
@@ -384,6 +520,19 @@ source_dirs([H|T], WD) -->
 	html(option([onClick(Call)|Attrs], H)),
 	source_dirs(T, WD).
 
+packs_link -->
+	{ pack_property(_,_), !,
+	  http_link_to_id(pldoc_pack, [], HREF),
+	  format(atom(Call), 'document.location=\'~w\';', [HREF])
+	},
+	html(option([ class(packs),
+		      onClick(Call),
+		      value(':packs:')
+		    ],
+		    'List extension packs')).
+packs_link -->
+	[].
+
 %%	source_directory(+Dir) is semidet.
 %%	source_directory(-Dir) is nondet.
 %
