@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@cs.vu.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2010, University of Amsterdam
+    Copyright (C): 1985-2013, University of Amsterdam
 			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
@@ -20,7 +18,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -34,9 +32,10 @@
 :- module(rdf_http_plugin, []).
 :- use_module(library(http/http_open)).
 :- use_module(library(http/http_header)).
-:- use_module(library(semweb/rdf_db)).
+:- use_module(library(semweb/rdf_db), []). % we define hooks for this
 :- use_module(library(date)).
 :- use_module(library(error)).
+:- use_module(library(lists)).
 
 
 /** <module> RDF HTTP Plugin
@@ -73,10 +72,10 @@ rdf_db:url_protocol(http).
 %	reply with the HTML description rather than the RDF).
 
 rdf_extra_headers(
-	[ request_header('Accept' = 'application/rdf+xml, \
-				     text/rdf+xml; q=0.9, \
-				     text/turtle, \
-				     application/x-turtle; q=0.8, \
+	[ request_header('Accept' = 'application/rdf+xml, \c
+				     text/rdf+xml; q=0.9, \c
+				     text/turtle, \c
+				     application/x-turtle; q=0.8, \c
 				     */*; q=0.1')
 	]).
 
@@ -152,17 +151,24 @@ major_content_type(ContentType, Major) :-
 	sub_atom(ContentType, 0, Pre, _, Major).
 major_content_type(Major, Major).
 
-%%	rdf_content_type(+ContentType, +URL)
+%%	rdf_content_type(+ContentType, -URL) is semidet.
 %
-%	Deduce the RDF encoding from the mime-type.
+%	Deduce the RDF encoding from the   mime-type.  This predicate is
+%	defined as multifile such that the user can associate additional
+%	content types to RDF formats.
 %
 %	@bug	The turtle parser only parses a subset of n3.
 
+:- multifile
+	rdf_content_type/2.
+
 rdf_content_type('text/rdf',		  xml).
+rdf_content_type('text/xml',		  xml).
 rdf_content_type('text/rdf+xml',	  xml).
 rdf_content_type('application/rdf+xml',	  xml).
 rdf_content_type('application/x-turtle',  turtle).
 rdf_content_type('application/turtle',	  turtle).
+rdf_content_type('application/trig',	  trig).
 rdf_content_type('text/turtle',		  turtle).
 rdf_content_type('text/rdf+n3',		  turtle).	% Bit dubious
 rdf_content_type('text/html',		  xhtml).
