@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2007, University of Amsterdam
+    Copyright (C): 2007-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -36,10 +35,12 @@
 	    rdf_list_library/0,
 	    rdf_list_library/1,		% +Ontology
 	    rdf_list_library/2,		% +Ontology, +Options
-	    rdf_library_index/2		% ?Id, ?Facet
+	    rdf_library_source/2,	% +Ontology, -SourceURL
+	    rdf_library_index/2,	% ?Id, ?Facet
+	    rdf_current_manifest/1	% -Manifest
 	  ]).
-:- use_module(library('semweb/rdf_db')).
-:- use_module(library('semweb/rdf_turtle')).
+:- use_module(library(semweb/rdf_db)).
+:- use_module(library(semweb/rdf_turtle)).
 :- use_module(library(rdf)).
 :- use_module(library(lists)).
 :- use_module(library(option)).
@@ -50,6 +51,22 @@
 :- use_module(library(uri)).
 :- use_module(library(http/http_open)).
 :- use_module(library(thread)).
+:- use_module(library(apply)).
+
+:- predicate_options(rdf_list_library/2, 2,
+		     [ indent(atom),
+		       show_graph(boolean),
+		       show_source(boolean),
+		       show_virtual(boolean)
+		     ]).
+:- predicate_options(rdf_load_library/2, 2,
+		     [ concurrent(positive_integer),
+		       import(boolean),
+		       load(boolean),
+		       base_uri(atom),
+		       claimed_source(atom),
+		       not_found(oneof([error,warning,silent]))
+		     ]).
 
 /** <module> RDF Library Manager
 
@@ -66,10 +83,13 @@ The typical usage scenario is
 ==
 
 @tbd	Add caching info
+@tbd	Allow using Manifests on HTTP servers
 @author Jan Wielemaker
 */
 
-:- rdf_register_ns(lib, 'http://www.swi-prolog.org/rdf/library/').
+:- rdf_register_ns(lib,  'http://www.swi-prolog.org/rdf/library/').
+:- rdf_register_ns(void, 'http://rdfs.org/ns/void#').
+:- rdf_register_ns(vann, 'http://purl.org/vocab/vann/').
 
 :- dynamic
 	manifest/2,			% Path, Time
@@ -121,12 +141,7 @@ rdf_load_library(Id) :-
 	rdf_load_library(Id, []).
 
 rdf_load_library(Id, Options) :-
-	load_commands(Id, Options, Pairs),
-	pairs_values(Pairs, Commands),
-	list_to_set(Commands, Cmds2),
-	delete_virtual(Cmds2, Cmds3),
-	find_conflicts(Cmds3),
-	check_existence(Cmds3, Cmds, Options),
+	cleaned_load_commands(Id, Cmds, Options),
 	(   option(concurrent(Threads), Options)
 	->  true
 	;   guess_concurrency(Cmds, Threads)
@@ -138,6 +153,30 @@ rdf_load_library(Id, Options) :-
 	;   true
 	).
 
+%%	rdf_library_source(+Id, -Source) is nondet.
+%
+%	True of Source is the URL that is  part of the given library Id.
+%	This predicate finds all indirect   dependencies.  It does _not_
+%	check whether the source exists or is valid.
+%
+%	@see uri_file_name/2 for converting file:// URLs to a filename.
+
+rdf_library_source(Id, Source) :-
+	cleaned_load_commands(Id, Cmds,
+			      [ import(true),
+				not_found(silent)
+			      ]),
+	member(rdf_load(Source, _), Cmds).
+
+
+cleaned_load_commands(Id, Cmds, Options) :-
+	load_commands(Id, Options, Pairs),
+	pairs_values(Pairs, Commands),
+	list_to_set(Commands, Cmds2),
+	delete_virtual(Cmds2, Cmds3),
+	find_conflicts(Cmds3),
+	check_existence(Cmds3, Cmds, Options).
+
 delete_virtual([], []).
 delete_virtual([virtual(_)|T0], T) :- !,
 	delete_virtual(T0, T).
@@ -219,7 +258,7 @@ check_existence(CommandsIn, Commands, Options) :-
 	option(not_found(Level), Options, error),
 	must_be(oneof([error,warning,silent]), Level),
 	(   Level == silent
-	->  true
+	->  Commands = CommandsIn
 	;   missing_urls(CommandsIn, Commands, Missing),
 	    (	Missing == []
 	    ->	true
@@ -270,11 +309,11 @@ guess_concurrency(Commands, Threads) :-
 count_uris([], 0, 0).
 count_uris([rdf_load(URL, _)|T], F, NF) :-
 	count_uris(T, F0, NF0),
-	(   sub_atom(URL, 0, _, _, 'file://')
-	->  F is F0 + 1,
-	    NF = NF0
-	;   NF is NF0 + 1,
+	(   web_url(URL)
+	->  NF is NF0 + 1,
 	    F = F0
+	;   F is F0 + 1,
+	    NF = NF0
 	).
 
 
@@ -301,15 +340,16 @@ dry_load(Id, Level, Options) :-
 	->  merge_base_uri(Facets, Options, Options1),
 	    merge_source(Facets, Options1, Options2),
 	    merge_blanks(Facets, Options2, Options3),
+	    merge_format(Facets, Options3, Options4),
 	    (   \+ memberchk(virtual, Facets)
-	    ->  load_options(Options3, File, RdfOptions),
+	    ->  load_options(Options4, File, RdfOptions),
 		assert(command(Level, rdf_load(File, RdfOptions)))
 	    ;	assert(command(Level, virtual(File)))
 	    ),
 	    (	option(import(true), Options, true)
 	    ->	Level1 is Level + 1,
-	        forall(member(imports(_, Import), Facets),
-		       import(Import, Level1, Options3))
+		forall(member(imports(Type, Import), Facets),
+		       import(Import, Level1, [type(Type)|Options4]))
 	    ;	true
 	    )
 	;   existence_error(ontology, Id)
@@ -317,25 +357,35 @@ dry_load(Id, Level, Options) :-
 
 merge_base_uri(Facets, Options0, Options) :-
 	(   option(base_uri(Base), Facets)
-	->  delete(Options0, base_uri(_), Options1),
+	->  exclude(name_option(base_uri), Options0, Options1),
 	    Options = [base_uri(Base)|Options1]
 	;   Options = Options0
 	).
 
 merge_source(Facets, Options0, Options) :-
 	(   option(claimed_source(Base), Facets)
-	->  delete(Options0, claimed_source(_), Options1),
+	->  exclude(name_option(claimed_source), Options0, Options1),
 	    Options = [claimed_source(Base)|Options1]
 	;   Options = Options0
 	).
 
 merge_blanks(Facets, Options0, Options) :-
 	(   option(blank_nodes(Share), Facets)
-	->  delete(Options0, blank_nodes(_), Options1),
+	->  exclude(name_option(blank_nodes), Options0, Options1),
 	    Options = [blank_nodes(Share)|Options1]
 	;   Options = Options0
 	).
 
+merge_format(Facets, Options0, Options) :-
+	(   option(format(Format), Facets)
+	->  exclude(name_option(format), Options0, Options1),
+	    Options = [format(Format)|Options1]
+	;   Options = Options0
+	).
+
+name_option(Name, Term) :-
+	functor(Term, Name, 1).
+
 load_options(Options, File, RDFOptions) :-
 	findall(O, load_option(Options, File, O), RDFOptions).
 
@@ -353,9 +403,15 @@ load_option(Options, File, base_uri(BaseURI)) :-
 	atom_concat(Base0, File, BaseURI).
 load_option(Options, _File, blank_nodes(Share)) :-
 	option(blank_nodes(Share), Options).
+load_option(Options, _File, format(Format)) :-
+	option(format(Format), Options).
 
 %%	import(+URL, +Level, +Options) is det.
 
+import(Path, Level, Options) :-
+	option(type(data_dump), Options), !,
+	load_options(Options, Path, RdfOptions),
+	assert(command(Level, rdf_load(Path, RdfOptions))).
 import(Path, Level, Options) :-
 	(   (   library(Id, Path, _)
 	    ->	true
@@ -448,7 +504,8 @@ existing_url_source(url(http, URL)) :- !,
 
 rdf_list_library :-
 	rdf_update_library_index,
-	(   rdf_library_index(Id, title(Title)),
+	(   rdf_library_index(Id, title(TitleLiteral)),
+	    plain_string(TitleLiteral, Title),
 	    format('~w ~t~20|~w', [Id, Title]),
 	    (	rdf_library_index(Id, version(Version))
 	    ->	format(' (version ~w)', [Version])
@@ -459,6 +516,10 @@ rdf_list_library :-
 	;   true
 	).
 
+plain_string(String, String) :-
+	atomic(String), !.
+plain_string(lang(en, String), String) :- !.
+plain_string(lang(_, String), String) :- !.
 
 %%	rdf_library_index(?Id, ?Facet) is nondet.
 %
@@ -494,6 +555,11 @@ rdf_list_library :-
 %		Share is one of =share= or =noshare=.  Default is to
 %		share.
 %
+%		* format(Format)
+%		Format of the resource.  Can be used to overrule
+%		if the format as derived from the HTTP content type
+%		is wrong.
+%
 %		* provides_ns(URL)
 %		Ontology provides definitions in the namespace URL.
 %		The formal definition of this is troublesome, but in
@@ -505,8 +571,8 @@ rdf_list_library :-
 %		means it contains triples that have predicates or
 %		objects in the given namespace.
 %
-%		* manifest(Path)
-%		Manifest file this ontology is defined in
+%		* manifest(URL)
+%		URL of the manifest in which this ontology is defined.
 %
 %		* virtual
 %		Entry is virtual (cannot be loaded)
@@ -575,8 +641,10 @@ attach_dir(Path, Visited) :-
 attach_dir(Path, Visited) :-
 	atom_concat(Path, '/*', Pattern),
 	expand_file_name(Pattern, Members),
-	(   member(Manifest, Members),
-	    is_manifest_file(Manifest)
+	(   manifest_file(MBase),
+	    rdf_extension(Ext),
+	    atomic_list_concat([Path, /, MBase, '.', Ext], Manifest),
+	    exists_file(Manifest)
 	->  process_manifest(Manifest)
 	;   print_message(silent, rdf(no_manifest(Path)))
 	),
@@ -600,10 +668,14 @@ hidden_base('cvs').			% Windows
 %	@param	Location is either a path name or a URL.
 
 process_manifest(Source) :-
-	(   uri_file_name(Source, Manifest0)
-	->  absolute_file_name(Manifest0, Manifest)
-	;   absolute_file_name(Source, Manifest)
-	),				% Manifest is a canonical filename
+	(   web_url(Source)
+	->  uri_normalized(Source, Manifest)
+	;   uri_file_name(Source, Manifest0)
+	->  absolute_file_name(Manifest0, ManifestFile),
+	    uri_file_name(Manifest, ManifestFile)
+	;   absolute_file_name(Source, ManifestFile),
+	    uri_file_name(Manifest, ManifestFile)
+	),				% Manifest is a canonical URI
 	source_time(Manifest, MT),
 	(   manifest(Manifest, Time),
 	    (	MT =< Time
@@ -637,6 +709,9 @@ process_triples(Manifest, Triples) :-
 extract_namespace(Triples, Mnemonic, Namespace) :-
 	edge(Triples, Decl, lib:mnemonic, literal(Mnemonic)),
 	edge(Triples, Decl, lib:namespace, Namespace).
+extract_namespace(Triples, Mnemonic, Namespace) :-
+	edge(Triples, Decl, vann:preferredNamespacePrefix, literal(Mnemonic)),
+	edge(Triples, Decl, vann:preferredNamespaceUri, literal(Namespace)).
 
 %%	extract_ontology(+Triples, -Ontology) is nondet.
 %
@@ -647,13 +722,27 @@ extract_ontology(Triples, library(Name, URL, Options)) :-
 	(   ontology_type(Type)
 	->  file_base_name(URL, BaseName),
 	    file_name_extension(Name, _, BaseName),
-	    findall(Facet, facet(Triples, URL, Facet), Options)
+	    findall(Facet, facet(Triples, URL, Facet), Options0),
+	    sort(Options0, Options1),
+	    keep_specialized_facets(Options1, Options)
 	).
 
+keep_specialized_facets(All, Special) :-
+	exclude(more_general(All), All, Special).
+
+more_general(All, Facet) :-
+	generalized(Facet, Special),
+	memberchk(Special, All).
+
+generalized(imports(ontology, Path), imports(Other, Path)) :-
+	dif(Other, ontology).
+
 ontology_type(X) :-
 	(   rdf_equal(X, lib:'Ontology')
 	;   rdf_equal(X, lib:'Schema')
 	;   rdf_equal(X, lib:'Instances')
+	;   rdf_equal(X, void:'Dataset')
+	;   rdf_equal(X, void:'Linkset')
 	).
 
 %%	facet(+Triples, +File, -Facet) is nondet.
@@ -661,12 +750,8 @@ ontology_type(X) :-
 %	Enumerate facets about File from   Triples. Facets are described
 %	with rdf_library_index/2.
 
-facet(Triples, File, title(Title)) :-
-	edge(Triples, File, dc:title, literal(Title)).
-:- if(rdf_current_ns(dcterms, _)).
 facet(Triples, File, title(Title)) :-
 	edge(Triples, File, dcterms:title, literal(Title)).
-:- endif.
 facet(Triples, File, version(Version)) :-
 	edge(Triples, File, owl:versionInfo, literal(Version)).
 facet(Triples, File, comment(Comment)) :-
@@ -675,6 +760,8 @@ facet(Triples, File, base_uri(BaseURI)) :-
 	edge(Triples, File, lib:baseURI, BaseURI).
 facet(Triples, File, claimed_source(Source)) :-
 	edge(Triples, File, lib:source, Source).
+facet(Triples, File, format(Format)) :-
+	edge(Triples, File, lib:format, literal(Format)).
 facet(Triples, File, blank_nodes(Mode)) :-
 	edge(Triples, File, lib:blankNodes, literal(Mode)),
 	must_be(oneof([share,noshare]), Mode).
@@ -684,6 +771,10 @@ facet(Triples, File, imports(schema, Path)) :-
 	edge(Triples, File, lib:schema, Path).
 facet(Triples, File, imports(instances, Path)) :-
 	edge(Triples, File, lib:instances, Path).
+facet(Triples, File, imports(subset, Path)) :-
+	edge(Triples, File, void:subset, Path).
+facet(Triples, File, imports(data_dump, Path)) :-
+	edge(Triples, File, void:dataDump, Path).
 facet(Triples, File, provides_ns(NS)) :-
 	edge(Triples, File, lib:providesNamespace, NSDecl),
 	edge(Triples, NSDecl, lib:namespace, NS).
@@ -691,14 +782,39 @@ facet(Triples, File, uses_ns(NS)) :-
 	edge(Triples, File, lib:usesNamespace, NSDecl),
 	edge(Triples, NSDecl, lib:namespace, NS).
 facet(Triples, File, virtual) :-
-	edge(Triples, File, rdf:type, lib:'Virtual').
+	(   edge(Triples, File, rdf:type, lib:'Virtual')
+	;   edge(Triples, File, rdf:type, void:'Dataset')
+	;   edge(Triples, File, rdf:type, void:'Linkset')
+	) -> true.
 
 %%	edge(+Triples, ?S, ?P, ?O) is nondet.
 %
-%	Like rdf/3 over a list of Triples.
+%	Like rdf_has/3 over a list of Triples.
 
 edge(Triples, S, P, O) :-
-	member(rdf(S,P,O), Triples).
+	nonvar(P), !,
+	sub_p(SubP, P),
+	member(rdf(S,SubP,O), Triples).
+edge(Triples, S, P, O) :-
+	member(rdf(S,SubP,O), Triples),
+	sub_p(SubP, P).
+
+sub_p(P, P).
+sub_p(Sub, P) :-
+	(   nonvar(Sub)
+	->  sub_property_of(Sub, Sub1),
+	    sub_p(Sub1, P)
+	;   sub_property_of(Sub1, P),
+	    sub_p(Sub, Sub1)
+	).
+
+:- rdf_meta
+	sub_property_of(r,r).
+
+sub_property_of(void:subset,	     owl:imports).
+sub_property_of(dcterms:description, rdfs:comment).
+sub_property_of(void:dataDump,	     owl:imports).
+sub_property_of(dc:title,	     dcterms:title).
 
 %%	source_time(+Source, -Modified) is semidet.
 %
@@ -707,7 +823,7 @@ edge(Triples, S, P, O) :-
 %	@error	existence_error(Type, Source).
 
 source_time(URL, Modified) :-
-	sub_atom(URL, 0, _, _, 'http://'), !,
+	web_url(URL), !,
 	http_open(URL, Stream,
 		  [ header(last_modified, Date),
 		    method(head)
@@ -721,29 +837,34 @@ source_time(URL, Modified) :-
 source_time(File, Modified) :-
 	time_file(File, Modified).
 
+web_url(URL) :-
+	sub_atom(URL, 0, _, _, 'http://').
 
-%%	read_triples(+File, -Triples) is det.
+
+%%	read_triples(+URL, -Triples) is det.
 %
 %	Read RDF/XML or Turtle file into a list of triples.
 
-read_triples(File, Triples) :-
-	file_name_extension(_, rdf, File), !,
-	load_rdf(File, Triples).
-read_triples(File, Triples) :-
-	file_name_extension(_, ttl, File), !,
-	rdf_load_turtle(File, Triples, []).
+read_triples(FileURL, Triples) :-
+	uri_file_name(FileURL, File), !,
+	(   file_name_extension(_, rdf, File)
+	->  load_rdf(File, Triples)
+	;   rdf_load_turtle(File, Triples, [])
+	).
+read_triples(HTTPURL, Triples) :-
+	file_name_extension(_, Ext, HTTPURL),
+	setup_call_cleanup(
+	    http_open(HTTPURL, In, []),
+	    stream_triples(In, Ext, Triples),
+	    close(In)).
 
-%%	is_manifest_file(+Path)
-%
-%	True if Path is the name of a manifest file.
+stream_triples(Stream, rdf, Triples) :-
+	load_rdf(stream(Stream), Triples).
+stream_triples(Stream, ttl, Triples) :-
+	rdf_load_turtle(stream(Stream), Triples, []).
 
-is_manifest_file(Path) :-
-	file_base_name(Path, File),
-	downcase_atom(File, Lwr),
-	file_name_extension(Base, Ext, Lwr),
-	manifest_file(Base),
-	rdf_extension(Ext), !.
 
+manifest_file('void').			% make order optional?
 manifest_file('Manifest').
 manifest_file('manifest').
 
@@ -803,6 +924,14 @@ define_namespace(ns(Mnemonic, Namespace)) :-
 			[
 			]).
 
+%%	rdf_current_manifest(-URL) is nondet.
+%
+%	True if URL is the URL of a currently loaded manifest file.
+
+rdf_current_manifest(URL) :-
+	manifest(URL, _Time).
+
+
 
 		 /*******************************
 		 *	      MESSAGES		*
