@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "pl-incl.h"
@@ -28,9 +28,6 @@
 #include "pl-codelist.h"
 #include <errno.h>
 #include <stdio.h>
-#ifdef __WINDOWS__
-#include "windows/mswchar.h"			/* Terrible hack */
-#endif
 #if HAVE_LIMITS_H
 #include <limits.h>			/* solaris compatibility */
 #endif
@@ -121,6 +118,42 @@ PL_from_stack_text(PL_chars_t *text)
 }
 
 
+#define INT64_DIGITS 20
+
+static char *
+ui64toa(uint64_t val, char *out)
+{ char tmpBuf[INT64_DIGITS + 1];
+  char *ptrOrg = tmpBuf + INT64_DIGITS;
+  char *ptr = ptrOrg;
+  size_t nbDigs;
+
+  do
+  { int rem = val % 10;
+
+    *--ptr = rem + '0';
+    val /= 10;
+  } while ( val );
+
+  nbDigs = ptrOrg - ptr;
+  memcpy(out, ptr, nbDigs);
+  out += nbDigs;
+  *out = '\0';
+
+  return out;				/* points to the END */
+};
+
+
+static char *
+i64toa(int64_t val, char *out)
+{ if ( val < 0 )
+  { *out++ = '-';
+    val = -val;
+  }
+
+  return ui64toa((uint64_t)val, out);
+}
+
+
 int
 PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
 { word w = valHandle(l);
@@ -138,11 +171,13 @@ PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
     PL_get_number(l, &n);
     switch(n.type)
     { case V_INTEGER:
-	sprintf(text->buf, INT64_FORMAT, n.value.i);
+      { char *ep = i64toa(n.value.i, text->buf);
+
         text->text.t    = text->buf;
-	text->length    = strlen(text->text.t);
+	text->length    = ep-text->text.t;
 	text->storage   = PL_CHARS_LOCAL;
 	break;
+      }
 #ifdef O_GMP
       case V_MPZ:
       { size_t sz = mpz_sizeinbase(n.value.mpz, 10) + 2;
@@ -197,7 +232,7 @@ PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
 	{ case CVT_partial:
 	    return PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
 	  case CVT_nolist:
-	    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, l);
+	    goto error;
 	  case CVT_nocode:
 	  case CVT_nochar:
 	  { term_t culprit = PL_new_term_ref();
@@ -237,9 +272,12 @@ PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
     encodings[1] = ENC_WCHAR;
     encodings[2] = ENC_UNKNOWN;
 
-    wflags = ((flags&CVT_WRITE_CANONICAL)
-		? PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS
-		: PL_WRT_NUMBERVARS);
+    if ( (flags&CVT_WRITEQ) == CVT_WRITEQ )
+      wflags = PL_WRT_QUOTED|PL_WRT_NUMBERVARS;
+    else if ( (flags&CVT_WRITE_CANONICAL) )
+      wflags = PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS;
+    else
+      wflags = PL_WRT_NUMBERVARS;
 
     for(enc = encodings; *enc != ENC_UNKNOWN; enc++)
     { size_t size;
@@ -292,7 +330,9 @@ error:
   if ( (flags & CVT_EXCEPTION) )
   { atom_t expected;
 
-    if ( flags & CVT_LIST )
+    if ( (flags & CVT_LIST) && !(flags&(CVT_ATOM|CVT_NUMBER)) )
+      expected = ATOM_list;		/* List and/or string object */
+    else if ( flags & CVT_LIST )
       expected = ATOM_text;
     else if ( flags & CVT_NUMBER )
       expected = ATOM_atomic;
@@ -495,12 +535,12 @@ PL_unify_text_range(term_t term, PL_chars_t *text,
     if ( offset > text->length || offset + len > text->length )
       return FALSE;
 
-    if ( len == 1 )
+    if ( len == 1 && type == PL_ATOM )
     { GET_LD
       int c;
 
       if ( text->encoding == ENC_ISO_LATIN_1 )
-	c = text->text.t[offset];
+	c = text->text.t[offset]&0xff;
       else
 	c = text->text.w[offset];
 
@@ -669,7 +709,7 @@ represented.
 
 static int
 wctobuffer(wchar_t c, mbstate_t *mbs, Buffer buf)
-{ char b[MB_LEN_MAX];
+{ char b[PL_MB_LEN_MAX];
   size_t n;
 
   if ( (n=wcrtomb(b, c, mbs)) != (size_t)-1 )
