@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -88,7 +87,7 @@ static void initHeapDebug(void);
 #define UNICODE_NOCHAR 0xFFFF
 #endif
 
-#if (_MSC_VER < 1400)
+#if (_MSC_VER < 1400) && !defined(__MINGW32__)
 typedef DWORD DWORD_PTR;
 #endif
 
@@ -182,6 +181,7 @@ static HICON    _rlc_hicon;		/* Global icon */
 
 static LRESULT WINAPI rlc_wnd_proc(HWND win, UINT msg, WPARAM wP, LPARAM lP);
 
+static void	rcl_setup_ansi_colors(RlcData b);
 static void	rlc_place_caret(RlcData b);
 static void	rlc_resize_pixel_units(RlcData b, int w, int h);
 static RlcData	rlc_make_buffer(int w, int h);
@@ -200,6 +200,8 @@ static int	rlc_add_lines(RlcData b, int here, int add);
 static void	rlc_start_selection(RlcData b, int x, int y);
 static void	rlc_extend_selection(RlcData b, int x, int y);
 static void	rlc_word_selection(RlcData b, int x, int y);
+static int	rlc_has_selection(RlcData b);
+static void	rlc_set_selection(RlcData b, int sl, int sc, int el, int ec);
 static void	rlc_copy(RlcData b);
 static void	rlc_destroy(RlcData b);
 static void	rlc_request_redraw(RlcData b);
@@ -207,7 +209,8 @@ static void	rlc_redraw(RlcData b);
 static int	rlc_breakargs(TCHAR *line, TCHAR **argv);
 static void	rlc_resize(RlcData b, int w, int h);
 static void	rlc_adjust_line(RlcData b, int line);
-static int	text_width(RlcData b, HDC hdc, const TCHAR *text, int len);
+static int	text_width(RlcData b, HDC hdc, const text_char *text, int len);
+static int	tchar_width(RlcData b, HDC hdc, const TCHAR *text, int len);
 static void	rlc_queryfont(RlcData b);
 static void     rlc_do_write(RlcData b, TCHAR *buf, int count);
 static void     rlc_reinit_line(RlcData b, int line);
@@ -216,10 +219,8 @@ static int	rlc_between(RlcData b, int f, int t, int v);
 static void	free_user_data(RlcData b);
 
 static RlcQueue	rlc_make_queue(int size);
-static void	rlc_free_queue(RlcQueue q);
 static int	rlc_from_queue(RlcQueue q);
 static int	rlc_is_empty_queue(RlcQueue q);
-static void	rlc_empty_queue(RlcQueue q);
 
 extern int	main();
 
@@ -389,7 +390,7 @@ rlc_main(HANDLE hInstance, HANDLE hPrevInstance,
 { TCHAR *	    argv[100];
   int		    argc;
   TCHAR		    program[MAXPATHLEN];
-  TCHAR	 	    progbase[100];
+  TCHAR		    progbase[100];
   RlcData           b;
   rlc_console_attr  attr;
 
@@ -487,7 +488,7 @@ rlc_create_window(RlcData b)
 		      NULL, NULL, _rlc_hinstance, NULL);
 
   b->window = hwnd;
-  SetWindowLong(hwnd, GWL_DATA, (LONG) b);
+  SetWindowLongPtr(hwnd, GWL_DATA, (LONG_PTR) b);
   SetScrollRange(hwnd, SB_VERT, 0, b->sb_lines, FALSE);
   SetScrollPos(hwnd, SB_VERT, b->sb_start, TRUE);
 
@@ -495,6 +496,7 @@ rlc_create_window(RlcData b)
   b->sb_lines = rlc_count_lines(b, b->first, b->last);
   b->sb_start = rlc_count_lines(b, b->first, b->window_start);
 
+  rcl_setup_ansi_colors(b);
   b->foreground = GetSysColor(COLOR_WINDOWTEXT);
   b->background = GetSysColor(COLOR_WINDOW);
   b->sel_foreground = GetSysColor(COLOR_HIGHLIGHTTEXT);
@@ -734,7 +736,7 @@ static HKEY
 rlc_option_key(rlc_console_attr *attr, int create)
 { TCHAR Prog[256];
   TCHAR *address[] = { _T("Software"),
-  		      RLC_VENDOR,
+		      RLC_VENDOR,
 		      Prog,
 		      _T("Console"),
 		      (TCHAR *)attr->key,	/* possible secondary key */
@@ -837,6 +839,59 @@ rlc_breakargs(TCHAR *line, TCHAR **argv)
 }
 
 
+		 /*******************************
+		 *	    ANSI COLORS		*
+		 *******************************/
+
+/* See http://en.wikipedia.org/wiki/ANSI_escape_code */
+
+static void
+rcl_setup_ansi_colors(RlcData b)
+{ b->sgr_flags = TF_DEFAULT;
+
+#ifdef ANSI_VGA_COLORS
+					/* normal versions */
+  b->ansi_color[0]  = RGB(  0,  0,  0);	/* black */
+  b->ansi_color[1]  = RGB(170,  0,  0);	/* red */
+  b->ansi_color[2]  = RGB(0,  170,  0);	/* green */
+  b->ansi_color[3]  = RGB(170, 85,  0);	/* yellow */
+  b->ansi_color[4]  = RGB(  0,  0,170);	/* blue */
+  b->ansi_color[5]  = RGB(170,  0,170);	/* magenta */
+  b->ansi_color[6]  = RGB(  0,170,170);	/* cyan */
+  b->ansi_color[7]  = RGB(170,170,170);	/* white */
+					/* bright/light versions */
+  b->ansi_color[8]  = RGB( 85, 85, 85);	/* black */
+  b->ansi_color[9]  = RGB(255, 85, 85);	/* red */
+  b->ansi_color[10] = RGB( 85,255, 85);	/* green */
+  b->ansi_color[11] = RGB(255,255, 85);	/* yellow */
+  b->ansi_color[12] = RGB( 85, 85,255);	/* blue */
+  b->ansi_color[13] = RGB(255, 85,255);	/* magenta */
+  b->ansi_color[14] = RGB( 85,255,255);	/* cyan */
+  b->ansi_color[15] = RGB(255,255,255);	/* white */
+#else /*XTERM*/
+					/* normal versions */
+  b->ansi_color[0]  = RGB(  0,  0,  0);	/* black */
+  b->ansi_color[1]  = RGB(205,  0,  0);	/* red */
+  b->ansi_color[2]  = RGB(0,  205,  0);	/* green */
+  b->ansi_color[3]  = RGB(205,205,  0);	/* yellow */
+  b->ansi_color[4]  = RGB(  0,  0,238);	/* blue */
+  b->ansi_color[5]  = RGB(205,  0,205);	/* magenta */
+  b->ansi_color[6]  = RGB(  0,205,205);	/* cyan */
+  b->ansi_color[7]  = RGB(229,229,229);	/* white */
+					/* bright/light versions */
+  b->ansi_color[8]  = RGB(127,127,127);	/* black */
+  b->ansi_color[9]  = RGB(255,  0,  0);	/* red */
+  b->ansi_color[10] = RGB(  0,255,  0);	/* green */
+  b->ansi_color[11] = RGB(255,255,  0);	/* yellow */
+  b->ansi_color[12] = RGB( 92, 92,255);	/* blue */
+  b->ansi_color[13] = RGB(255,  0,255);	/* magenta */
+  b->ansi_color[14] = RGB(  0,255,255);	/* cyan */
+  b->ansi_color[15] = RGB(255,255,255);	/* white */
+#endif
+}
+
+
+
 		 /*******************************
 		 *	     ATTRIBUTES		*
 		 *******************************/
@@ -931,7 +986,14 @@ rlc_interrupt(RlcData b)
 
 static void
 typed_char(RlcData b, int chr)
-{ if ( chr == Control('C') )
+{ if ( chr == Control('C') && rlc_has_selection(b) )
+  { rlc_copy(b);
+    return;
+  }
+
+  rlc_set_selection(b, 0, 0, 0, 0);
+
+  if ( chr == Control('C') )
     rlc_interrupt(b);
   else if ( chr == Control('V') || chr == Control('Y') )
     rlc_paste(b);
@@ -964,7 +1026,7 @@ IsDownKey(code)
 
 static LRESULT WINAPI
 rlc_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
-{ RlcData b = (RlcData) GetWindowLong(hwnd, GWL_DATA);
+{ RlcData b = (RlcData) GetWindowLongPtr(hwnd, GWL_DATA);
 
   switch(message)
   { case WM_CREATE:
@@ -985,7 +1047,7 @@ rlc_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 
       if ( placement.showCmd == SW_SHOWNORMAL )
       { b->win_x = placement.rcNormalPosition.left;
-  	b->win_y = placement.rcNormalPosition.top;
+	b->win_y = placement.rcNormalPosition.top;
 
 	b->modified_options |= OPT_POSITION;
       }
@@ -1019,6 +1081,7 @@ rlc_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	  rlc_paste(b);
 	  return 0;
 	case IDM_COPY:
+	  rlc_copy(b);
 	  return 0;			/* no op: already done */
 	case IDM_CUT:
 	  break;			/* TBD: cut */
@@ -1056,6 +1119,7 @@ rlc_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	case VK_DOWN:	chr = Control('N');	break;
 	case VK_HOME:	chr = Control('A');	break;
 	case VK_END:	chr = Control('E');	break;
+        case VK_CANCEL: rlc_interrupt(b);       return 0;
 
         case VK_PRIOR:			/* page up */
 	{ int maxdo = rlc_count_lines(b, b->first, b->window_start);
@@ -1411,7 +1475,7 @@ rlc_is_word_char(int chr)
 		 *	    SELECTION		*
 		 *******************************/
 
-#define SelLT(l1, c1, l2, c2) ((l1) < (l2) || (l1) == (l2) && (c1) < (c2))
+#define SelLT(l1, c1, l2, c2) ((l1) < (l2) || ((l1) == (l2) && (c1) < (c2)))
 #define SelEQ(l1, c1, l2, c2) ((l1) == (l2) && (c1) == (c2))
 
 static int
@@ -1521,7 +1585,7 @@ rlc_translate_mouse(RlcData b, int x, int y, int *line, int *chr)
   } else if ( tl->size == 0 )
   { *chr = 0;
   } else
-  { TCHAR *s = tl->text;
+  { text_char *s = tl->text;
     HDC hdc = GetDC(b->window);
     int f = 0;
     int t = tl->size;
@@ -1537,7 +1601,7 @@ rlc_translate_mouse(RlcData b, int x, int y, int *line, int *chr)
       if ( x > w )
       { int cw;
 
-	GetCharWidth32(hdc, s[m], s[m], &cw);
+	GetCharWidth32(hdc, s[m].code, s[m].code, &cw);
 	if ( x < w+cw )
 	{ *chr = m;
 	  return;
@@ -1565,19 +1629,6 @@ rlc_start_selection(RlcData b, int x, int y)
 }
 
 
-static void
-rlc_end_selection(RlcData b, int x, int y)
-{ int l, c;
-
-  rlc_translate_mouse(b, x, y, &l, &c);
-  if ( SelLT(l, c, b->sel_org_line, b->sel_org_char) )
-    rlc_set_selection(b, l, c, b->sel_org_line, b->sel_org_char);
-  else if ( SelLT(b->sel_org_line, b->sel_org_char, l, c) )
-    rlc_set_selection(b, b->sel_org_line, b->sel_org_char, l, c);
-  rlc_set_selection(b, l, c, l, c);
-}
-
-
 static int				/* v >= f && v <= t */
 rlc_between(RlcData b, int f, int t, int v)
 { int h = rlc_count_lines(b, b->first, v);
@@ -1598,12 +1649,12 @@ rlc_word_selection(RlcData b, int x, int y)
   if ( rlc_between(b, b->first, b->last, l) )
   { TextLine tl = &b->lines[l];
 
-    if ( c < tl->size && rlc_is_word_char(tl->text[c]) )
+    if ( c < tl->size && rlc_is_word_char(tl->text[c].code) )
     { int f, t;
 
-      for(f=c; f>0 && rlc_is_word_char(tl->text[f-1]); f--)
+      for(f=c; f>0 && rlc_is_word_char(tl->text[f-1].code); f--)
 	;
-      for(t=c; t<tl->size && rlc_is_word_char(tl->text[t]); t++)
+      for(t=c; t<tl->size && rlc_is_word_char(tl->text[t].code); t++)
 	;
       rlc_set_selection(b, l, f, l, t);
     }
@@ -1625,15 +1676,15 @@ rlc_extend_selection(RlcData b, int x, int y)
     { if ( rlc_between(b, b->first, b->last, l) )
       { TextLine tl = &b->lines[l];
 
-	if ( c < tl->size && rlc_is_word_char(tl->text[c]) )
-	  for(; c > 0 && rlc_is_word_char(tl->text[c-1]); c--)
+	if ( c < tl->size && rlc_is_word_char(tl->text[c].code) )
+	  for(; c > 0 && rlc_is_word_char(tl->text[c-1].code); c--)
 	    ;
       }
       if ( rlc_between(b, b->first, b->last, el) )
       { TextLine tl = &b->lines[el];
 
-	if ( ec < tl->size && rlc_is_word_char(tl->text[ec]) )
-	  for(; ec < tl->size && rlc_is_word_char(tl->text[ec]); ec++)
+	if ( ec < tl->size && rlc_is_word_char(tl->text[ec].code) )
+	  for(; ec < tl->size && rlc_is_word_char(tl->text[ec].code); ec++)
 	    ;
       }
     } else if ( b->sel_unit == SEL_LINE )
@@ -1644,15 +1695,15 @@ rlc_extend_selection(RlcData b, int x, int y)
     { if ( rlc_between(b, b->first, b->last, l) )
       { TextLine tl = &b->lines[l];
 
-	if ( c < tl->size && rlc_is_word_char(tl->text[c]) )
-	  for(; c < tl->size && rlc_is_word_char(tl->text[c]); c++)
+	if ( c < tl->size && rlc_is_word_char(tl->text[c].code) )
+	  for(; c < tl->size && rlc_is_word_char(tl->text[c].code); c++)
 	    ;
       }
       if ( rlc_between(b, b->first, b->last, el) )
       { TextLine tl = &b->lines[el];
 
-	if ( ec < tl->size && rlc_is_word_char(tl->text[ec]) )
-	  for(; ec > 0 && rlc_is_word_char(tl->text[ec-1]); ec--)
+	if ( ec < tl->size && rlc_is_word_char(tl->text[ec].code) )
+	  for(; ec > 0 && rlc_is_word_char(tl->text[ec-1].code); ec--)
 	    ;
       }
     } else if ( b->sel_unit == SEL_LINE )
@@ -1668,7 +1719,7 @@ rlc_read_from_window(RlcData b, int sl, int sc, int el, int ec)
   TCHAR *buf;
   int i = 0;
 
-  if ( el < sl || el == sl && ec < sc )
+  if ( el < sl || (el == sl && ec < sc) )
     return NULL;			/* invalid region */
   if ( !(buf = rlc_malloc(bufsize * sizeof(TCHAR))) )
     return NULL;			/* not enough memory */
@@ -1682,22 +1733,22 @@ rlc_read_from_window(RlcData b, int sl, int sc, int el, int ec)
 	e = tl->size;
 
       while(sc < e)
-      { if ( i >= bufsize )
+      { if ( i+1 >= bufsize )
 	{ bufsize *= 2;
 	  if ( !(buf = rlc_realloc(buf, bufsize * sizeof(TCHAR))) )
 	    return NULL;		/* not enough memory */
 	}
-	buf[i++] = tl->text[sc++];
+	buf[i++] = tl->text[sc++].code;
       }
     }
 
     if ( sl == el || sl == b->last )
-    { buf[i++] = '\0';
+    { buf[i++] = '\0';			/* Always room for the 0 */
       return buf;
     }
 
     if ( tl && !tl->softreturn )
-    { if ( i+1 >= bufsize )
+    { if ( i+2 >= bufsize )
       { bufsize *= 2;
 	if ( !(buf = rlc_realloc(buf, bufsize * sizeof(TCHAR))) )
 	  return NULL;			/* not enough memory */
@@ -1709,15 +1760,22 @@ rlc_read_from_window(RlcData b, int sl, int sc, int el, int ec)
 }
 
 
-static TCHAR *
-rlc_selection(RlcData b)
+static int
+rlc_has_selection(RlcData b)
 { if ( SelEQ(b->sel_start_line, b->sel_start_char,
 	     b->sel_end_line,   b->sel_end_char) )
-    return NULL;
+    return FALSE;
+  return TRUE;
+}
 
-  return rlc_read_from_window(b,
-			      b->sel_start_line, b->sel_start_char,
-			      b->sel_end_line,   b->sel_end_char);
+
+static TCHAR *
+rlc_selection(RlcData b)
+{ if ( rlc_has_selection(b) )
+    return rlc_read_from_window(b,
+				b->sel_start_line, b->sel_start_char,
+				b->sel_end_line,   b->sel_end_char);
+  return NULL;
 }
 
 
@@ -1771,16 +1829,16 @@ rlc_place_caret(RlcData b)
       { SetCaretPos((b->caret_x + 1) * b->cw, line * b->ch);
       } else
       { HDC hdc = GetDC(b->window);
-	SIZE tsize;
+	int tw;
 	TextLine tl = &b->lines[b->caret_y];
 	HFONT old;
 
 	old = SelectObject(hdc, b->hfont);
-	GetTextExtentPoint32(hdc, tl->text, b->caret_x, &tsize);
+	tw = text_width(b, hdc, tl->text, b->caret_x);
 	SelectObject(hdc, old);
 	ReleaseDC(b->window, hdc);
 
-	SetCaretPos(b->cw + tsize.cx, line * b->ch);
+	SetCaretPos(b->cw + tw, line * b->ch);
       }
       if ( !b->caret_is_shown )
       { ShowCaret(b->window);
@@ -1818,6 +1876,80 @@ rlc_update_scrollbar(RlcData b)
 }
 
 
+static void
+rcl_paint_text(RlcData b, HDC hdc,
+	       TextLine tl, int from, int to,
+	       int ty, int *cx, int insel)
+{ text_char *chars, *s;
+  text_char buf[MAXLINE];
+  TCHAR text[MAXLINE];
+  TCHAR *t;
+  int len = to-from;
+  int i;
+
+  if ( len <= 0 )
+    return;
+
+  if ( tl->text && to <= tl->size )
+  { chars = &tl->text[from];
+  } else
+  { text_char *o;
+    int copy;
+
+    o = chars = buf;
+    s = &tl->text[from];
+    copy = tl->text ? tl->size-from : 0;
+    for(i=0; i<copy; i++)
+      *o++ = *s++;
+    for(; i<len; i++, o++)
+    { o->code = ' ';
+      o->flags = TF_DEFAULT;
+    }
+  }
+
+  for(t=text, s=chars, i=0; i < len; i++, t++, s++)
+    *t = s->code;
+
+  if ( insel )					/* TBD: Cache */
+  { SetBkColor(hdc, b->sel_background);
+    SetTextColor(hdc, b->sel_foreground);
+    TextOut(hdc, *cx, ty, text, len);
+    *cx += tchar_width(b, hdc, text, len);
+  } else
+  { int start, segment;
+
+    for(start=0, s=chars, t=text;
+	start<len;
+	start+=segment, s+=segment, t+=segment)
+    { text_flags flags = s->flags;
+      int left = len-start;
+
+      for(segment=0; s[segment].flags == flags && segment<left; segment++)
+	;
+
+      if ( TF_FG(flags) == ANSI_COLOR_DEFAULT )
+	SetTextColor(hdc, b->foreground);
+      else
+	SetTextColor(hdc, b->ansi_color[TF_FG(flags)]);
+
+      if ( TF_BG(flags) == ANSI_COLOR_DEFAULT )
+	SetBkColor(hdc, b->background);
+      else
+	SetBkColor(hdc, b->ansi_color[TF_BG(flags)]);
+
+      TextOut(hdc, *cx, ty, t, segment);
+      if ( TF_BOLD(flags) )
+      { SetBkMode(hdc, TRANSPARENT);
+	TextOut(hdc, (*cx)+1, ty, t, segment);
+	TextOut(hdc, *cx, ty+1, t, segment);
+	SetBkMode(hdc, OPAQUE);
+      }
+      *cx += tchar_width(b, hdc, t, segment);
+    }
+  }
+}
+
+
 static void
 rlc_redraw(RlcData b)
 { PAINTSTRUCT ps;
@@ -1861,27 +1993,9 @@ rlc_redraw(RlcData b)
 
   for(; pl <= el; l = NextLine(b, l), pl++)
   { TextLine tl = &b->lines[l];
-    TCHAR text[MAXLINE];
     int ty = b->ch * pl;
     int cx = b->cw;
 
-    if ( !tl->text )
-    { int i;
-      TCHAR *t;
-
-      tl->size = 0;
-      for(i=0, t=text; i<b->width; i++)
-	*t++ = ' ';
-    } else
-    { int i;
-      TCHAR *t, *s;
-
-      for(i=0, t=text, s=tl->text; i<tl->size; i++)
-	*t++ = *s++;
-      for(; i<b->width; i++)
-	*t++ = ' ';
-    }
-
     rect.top    = ty;
     rect.bottom = rect.top + b->ch;
 
@@ -1890,34 +2004,23 @@ rlc_redraw(RlcData b)
     { int cf = b->sel_start_char;
       int ce = (b->sel_end_line != b->sel_start_line ? b->width
 						     : b->sel_end_char);
-      if ( cf > 0 )
-      {	TextOut(hdc, cx, ty, text, cf);
-	cx += text_width(b, hdc, text, cf);
-      }
-      SetBkColor(hdc, b->sel_background);
-      SetTextColor(hdc, b->sel_foreground);
-      TextOut(hdc, cx, ty, &text[cf], ce-cf);
-      cx += text_width(b, hdc, &text[cf], ce-cf);
+
+      rcl_paint_text(b, hdc, tl,  0, cf, ty, &cx, insel);
+      insel = TRUE;
+      rcl_paint_text(b, hdc, tl, cf, ce, ty, &cx, insel);
       if ( l == b->sel_end_line )
-      { SetBkColor(hdc, b->background);
-	SetTextColor(hdc, b->foreground);
-	TextOut(hdc, cx, ty, &text[ce], b->width - ce);
-	cx += text_width(b, hdc, &text[ce], b->width - ce);
+      { insel = FALSE;
+	rcl_paint_text(b, hdc, tl, ce, b->width, ty, &cx, insel);
       } else
 	insel = TRUE;
     } else if ( l == b->sel_end_line )	/* end of selection */
     { int ce = b->sel_end_char;
 
+      rcl_paint_text(b, hdc, tl, 0, ce, ty, &cx, insel);
       insel = FALSE;
-      TextOut(hdc, cx, ty, text, ce);
-      cx += text_width(b, hdc, text, ce);
-      SetBkColor(hdc, b->background);
-      SetTextColor(hdc, b->foreground);
-      TextOut(hdc, cx, ty, &text[ce], b->width - ce);
-      cx += text_width(b, hdc, &text[ce], b->width - ce);
+      rcl_paint_text(b, hdc, tl, ce, b->width, ty, &cx, insel);
     } else				/* entire line in/out selection */
-    { TextOut(hdc, cx, ty, text, b->width);
-      cx += text_width(b, hdc, text, b->width);
+    { rcl_paint_text(b, hdc, tl, 0, b->width, ty, &cx, insel);
     }
 
 					/* clear remainder of line */
@@ -1963,7 +2066,6 @@ rlc_request_redraw(RlcData b)
     int y = 0;
     RECT rect;
     int first = TRUE;
-    int clear = FALSE;
 
     rect.left = b->cw;
     rect.right = (b->width+1) * b->cw;
@@ -1978,16 +2080,13 @@ rlc_request_redraw(RlcData b)
 	  first = FALSE;
 	} else
 	  rect.bottom = (y+1) * b->ch;
-
-	if ( l->changed & CHG_CLEAR )
-	  clear = TRUE;
       }
       if ( i == b->last )
 	break;
     }
 
     if ( !first && b->window )
-      InvalidateRect(b->window, &rect, FALSE); /*clear);*/
+      InvalidateRect(b->window, &rect, FALSE);
     else if ( b->changed & CHG_CARET )
       rlc_place_caret(b);
   }
@@ -2086,7 +2185,25 @@ rlc_init_text_dimensions(RlcData b, HFONT font)
 
 
 static int
-text_width(RlcData b, HDC hdc, const TCHAR *text, int len)
+text_width(RlcData b, HDC hdc, const text_char *text, int len)
+{ if ( b->fixedfont )
+  { return len * b->cw;
+  } else
+  { SIZE size;
+    TCHAR tmp[MAXLINE];
+    int i;
+
+    for(i=0; i<len; i++)
+      tmp[i] = text[i].code;
+
+    GetTextExtentPoint32(hdc, tmp, len, &size);
+    return size.cx;
+  }
+}
+
+
+static int
+tchar_width(RlcData b, HDC hdc, const TCHAR *text, int len)
 { if ( b->fixedfont )
   { return len * b->cw;
   } else
@@ -2137,7 +2254,7 @@ rlc_queryfont(RlcData b)
   cf.hwndOwner   = b->window;
   cf.lpLogFont   = &lf;
   cf.Flags       = CF_SCREENFONTS|
-    		   CF_NOVERTFONTS|
+		   CF_NOVERTFONTS|
 		   CF_NOSIMULATIONS|
 		   CF_FORCEFONTEXIST|
 		   CF_INITTOLOGFONTSTRUCT;
@@ -2175,8 +2292,8 @@ rlc_make_buffer(int w, int h)
   b->changed	    = CHG_CARET|CHG_CHANGED|CHG_CLEAR;
   b->imode	    = IMODE_COOKED;	/* switch on first rlc_read() call */
   b->imodeswitch    = FALSE;
-  b->lhead 	    = NULL;
-  b->ltail 	    = NULL;
+  b->lhead	    = NULL;
+  b->ltail	    = NULL;
 
   memset(b->lines, 0, sizeof(text_line) * h);
   for(i=0; i<h; i++)
@@ -2257,15 +2374,15 @@ rlc_resize(RlcData b, int w, int h)
 	DEBUG(Dprint_lines(b, b->first, b->first));
 	DEBUG(Dprintf(_T("b->first = %d, b->last = %d\n"), b->first, b->last));
 	pl = &b->lines[PrevLine(b, i)];	/* this is the moved line */
-	tl->text = rlc_malloc((pl->size - w)*sizeof(TCHAR));
-	memmove(tl->text, &pl->text[w], (pl->size - w)*sizeof(TCHAR));
+	tl->text = rlc_malloc((pl->size - w)*sizeof(text_char));
+	memmove(tl->text, &pl->text[w], (pl->size - w)*sizeof(text_char));
 	DEBUG(Dprintf(_T("Copied %d chars from line %d to %d\n"),
 		      pl->size - w, pl - b->lines, i));
 	tl->size = pl->size - w;
 	tl->adjusted = TRUE;
 	tl->softreturn = FALSE;
 	pl->softreturn = TRUE;
-	pl->text = rlc_realloc(pl->text, w * sizeof(TCHAR));
+	pl->text = rlc_realloc(pl->text, w * sizeof(text_char));
 	pl->size = w;
 	pl->adjusted = TRUE;
 	i = (int)(pl - b->lines);
@@ -2277,9 +2394,9 @@ rlc_resize(RlcData b, int w, int h)
 	if ( i == b->last )
 	  rlc_add_line(b);
 	nl = &b->lines[NextLine(b, i)];
-	nl->text = rlc_realloc(nl->text, (nl->size + move)*sizeof(TCHAR));
-	memmove(&nl->text[move], nl->text, nl->size*sizeof(TCHAR));
-	memmove(nl->text, &tl->text[w], move*sizeof(TCHAR));
+	nl->text = rlc_realloc(nl->text, (nl->size + move)*sizeof(text_char));
+	memmove(&nl->text[move], nl->text, nl->size*sizeof(text_char));
+	memmove(nl->text, &tl->text[w], move*sizeof(text_char));
 	nl->size += move;
 	tl->size = w;
       }
@@ -2290,9 +2407,9 @@ rlc_resize(RlcData b, int w, int h)
 	rlc_add_line(b);
       nl = &b->lines[NextLine(b, i)];
 
-      nl->text = rlc_realloc(nl->text, (nl->size + tl->size)*sizeof(TCHAR));
-      memmove(&nl->text[tl->size], nl->text, nl->size*sizeof(TCHAR));
-      memmove(nl->text, tl->text, tl->size*sizeof(TCHAR));
+      nl->text = rlc_realloc(nl->text, (nl->size + tl->size)*sizeof(text_char));
+      memmove(&nl->text[tl->size], nl->text, nl->size*sizeof(text_char));
+      memmove(nl->text, tl->text, tl->size*sizeof(text_char));
       nl->size += tl->size;
       nl->adjusted = TRUE;
       rlc_shift_lines_up(b, i);
@@ -2346,8 +2463,8 @@ rlc_adjust_line(RlcData b, int line)
 
   if ( tl->text && !tl->adjusted )
   { tl->text = rlc_realloc(tl->text, tl->size == 0
-			   	? sizeof(TCHAR)
-				: tl->size * sizeof(TCHAR));
+				? sizeof(text_char)
+				: tl->size * sizeof(text_char));
     tl->adjusted = TRUE;
   }
 }
@@ -2359,11 +2476,11 @@ rlc_unadjust_line(RlcData b, int line)
 
   if ( tl->text )
   { if ( tl->adjusted )
-    { tl->text = rlc_realloc(tl->text, (b->width + 1)*sizeof(TCHAR));
+    { tl->text = rlc_realloc(tl->text, (b->width + 1)*sizeof(text_char));
       tl->adjusted = FALSE;
     }
   } else
-  { tl->text = rlc_malloc((b->width + 1)*sizeof(TCHAR));
+  { tl->text = rlc_malloc((b->width + 1)*sizeof(text_char));
     tl->adjusted = FALSE;
     tl->size = 0;
   }
@@ -2381,7 +2498,7 @@ rlc_open_line(RlcData b)
     b->first = NextLine(b, b->first);
   }
 
-  b->lines[i].text       = rlc_malloc((b->width + 1)*sizeof(TCHAR));
+  b->lines[i].text       = rlc_malloc((b->width + 1)*sizeof(text_char));
   b->lines[i].adjusted   = FALSE;
   b->lines[i].size       = 0;
   b->lines[i].softreturn = FALSE;
@@ -2507,7 +2624,11 @@ rlc_tab(RlcData b)
   { rlc_unadjust_line(b, b->caret_y);
 
     while ( tl->size < b->caret_x )
-      tl->text[tl->size++] = ' ';
+    { text_char *tc = &tl->text[tl->size++];
+
+      tc->code = ' ';
+      tc->flags = b->sgr_flags;
+    }
   }
 
   b->changed |= CHG_CARET;
@@ -2569,14 +2690,45 @@ rlc_erase_line(RlcData b)
 }
 
 
+static void
+rlc_sgr(RlcData b, int sgr)
+{ if ( sgr == 0 )
+  { b->sgr_flags = TF_DEFAULT;
+  } else if ( sgr >= 30 && sgr <= 39 )
+  { b->sgr_flags = TF_SET_FG(b->sgr_flags,
+			     sgr == 39 ? ANSI_COLOR_DEFAULT : sgr-30);
+  } else if ( sgr >= 40 && sgr <= 49 )
+  { b->sgr_flags = TF_SET_BG(b->sgr_flags,
+			     sgr == 49 ? ANSI_COLOR_DEFAULT : sgr-40);
+  } else if ( sgr >= 90 && sgr <= 99 )
+  { b->sgr_flags = TF_SET_FG(b->sgr_flags,
+			     sgr == 99 ? ANSI_COLOR_DEFAULT : sgr-90+8);
+  } else if ( sgr >= 100 && sgr <= 109 )
+  { b->sgr_flags = TF_SET_BG(b->sgr_flags,
+			     sgr == 109 ? ANSI_COLOR_DEFAULT : sgr-100+8);
+  } else if ( sgr == 1 )
+  { b->sgr_flags = TF_SET_BOLD(b->sgr_flags, 1);
+  } else if ( sgr == 4 )
+  { b->sgr_flags = TF_SET_UNDERLINE(b->sgr_flags, 1);
+  }
+}
+
+
 static void
 rlc_put(RlcData b, int chr)
 { TextLine tl = &b->lines[b->caret_y];
+  text_char *tc;
 
   rlc_unadjust_line(b, b->caret_y);
   while( tl->size < b->caret_x )
-    tl->text[tl->size++] = ' ';
-  tl->text[b->caret_x] = chr;
+  { tc = &tl->text[tl->size++];
+
+    tc->code  = ' ';
+    tc->flags = b->sgr_flags;
+  }
+  tc = &tl->text[b->caret_x];
+  tc->code = chr;
+  tc->flags = b->sgr_flags;
   if ( tl->size <= b->caret_x )
     tl->size = b->caret_x + 1;
   tl->changed |= CHG_CHANGED;
@@ -2694,6 +2846,14 @@ rlc_putansi(RlcData b, int chr)
 	case 'K':
 	  CMD(rlc_erase_line(b));
 	  break;
+	case 'm':
+	  { int i;
+	    rlc_need_arg(b, 1, 0);
+
+	    for(i=0; i<b->argc; i++)
+	      CMD(rlc_sgr(b, b->argv[i]));
+	    break;
+	  }
       }
       b->cmdstat = CMD_INITIAL;
   }
@@ -2893,7 +3053,7 @@ window_loop(LPVOID arg)
   if ( b->closing <= 2 )
   { MSG msg;
     TCHAR *waiting = _T("\r\nWaiting for Prolog. ")
-      		     _T("Close again to force termination ..");
+		     _T("Close again to force termination ..");
 
     rlc_write(b, waiting, _tcslen(waiting));
 
@@ -3048,16 +3208,6 @@ rlc_make_queue(int size)
 }
 
 
-void
-rlc_free_queue(RlcQueue q)
-{ if ( q )
-  { if ( q->buffer )
-      rlc_free(q->buffer);
-    rlc_free(q);
-  }
-}
-
-
 static int
 rlc_resize_queue(RlcQueue q, int size)
 { TCHAR *newbuf;
@@ -3114,12 +3264,6 @@ rlc_is_empty_queue(RlcQueue q)
 }
 
 
-void
-rlc_empty_queue(RlcQueue q)
-{ q->first = q->last = 0;
-}
-
-
 static int
 rlc_from_queue(RlcQueue q)
 { if ( q->first != q->last )
@@ -3370,8 +3514,6 @@ rlc_clearprompt(rlc_console c)
 		 *	    MISC STUFF		*
 		 *******************************/
 
-static TCHAR current_title[RLC_TITLE_MAX];
-
 void
 rlc_title(rlc_console c, TCHAR *title, TCHAR *old, int size)
 { RlcData b = rlc_get_data(c);
@@ -3391,11 +3533,7 @@ rlc_title(rlc_console c, TCHAR *title, TCHAR *old, int size)
 void
 rlc_icon(rlc_console c, HICON icon)
 {
-#ifdef WIN64
-  SetClassLong(rlc_hwnd(c), GCLP_HICON, (LONG) icon);
-#else
-  SetClassLong(rlc_hwnd(c), GCL_HICON, (LONG) icon);
-#endif
+  SetClassLongPtr(rlc_hwnd(c), GCLP_HICON, (LONG_PTR) icon);
 }
 
 
@@ -3650,7 +3788,7 @@ Dprintf(const TCHAR *fmt, ...)
   va_list args;
 
   va_start(args, fmt);
-  _vstprintf(buf, fmt, args);
+  vswprintf(buf, sizeof(buf)/sizeof(TCHAR), fmt, args);
   va_end(args);
 
   OutputDebugString(buf);
